/*
 * Author und Copyright: Ruediger Berlich
 *
 * Distributed under the Boost Software License, Version 1.0. 
 * See http://www.boost.org/LICENSE_1_0.txt
 *
 * Run with the command "root -l erzeuger.cpp" .
 * Requires a ROOT-version >= 6
 * Tested with ROOT 6.02.02 on MacOS X Yosemite
 */

#include <memory>
#include <functional>
#include <random>
#include <tuple>
#include <cmath>
#include <iostream>

typedef std::mt19937 GENERATOR;

/**
 * Retrieve a random number that is distributed according to a given function.
 * No error checks are done, for reasons of simplicity.
 */
std::tuple<double, double> getRandomNumberWithDistribution (
  const std::function<double(double)>& f
  , GENERATOR& generator
  , double x_min, double x_max
  , double y_min, double y_max
) {
  double x=0., y=0.;
  
  // The random distributions
  std::uniform_real_distribution<double> distribution_x(x_min, x_max);
  std::uniform_real_distribution<double> distribution_y(y_min, y_max);
  
  // Retrieve random variables in the allowed x- and y-range until y is < f(x)
  do {
    x = distribution_x(generator);
    y = distribution_y(generator);
  } while(f(x) < y);
  
  return std::make_tuple(x,y);
}

using namespace std::placeholders;

const std::size_t NENTRIES=10000;
const double x_min = -2*M_PI, x_max = 2*M_PI;
const double y_min = -1., y_max = 1.5;

/**
 * The "main()"-function (for ROOT: A function with the same name as the file)
 */
void erzeuger(void) {
  // Will hold the data
  std::vector<double> x_arr, y_arr;

  // Create a C++11-style random number generator.
  std::random_device rand_dev;          // Needed to get a good random seed
  GENERATOR generator(rand_dev());

  // Prepare the graphical output: Create a canvas
  TCanvas *cc = new TCanvas("cc","cc",0,0,1024,1024);
  
  // The random number distribution
  std::function<double(double)> sinus_f = std::bind<double(double)>(sin, _1);

  // Prepare the data for the graph
  for(std::size_t i=0; i<NENTRIES; i++) {
    auto data = getRandomNumberWithDistribution(
      sinus_f
      , generator
      , x_min, x_max
      , y_min, y_max
    );

    x_arr.push_back(std::get<0>(data));
    y_arr.push_back(std::get<1>(data));
  }

  // Pepare a TGraph object to hold and plot the data
  TGraph *sine_graph = new TGraph(NENTRIES, &x_arr[0], &y_arr[0]);

  // Set some plot properties
  sine_graph->SetTitle("Random numbers with sine-distribution, sine-function");
  sine_graph->SetMarkerStyle(21);
  sine_graph->SetMarkerSize(0.35);
  sine_graph->SetMarkerColor(16);
  sine_graph->GetXaxis()->SetLimits(x_min, x_max);
  sine_graph->GetXaxis()->SetTitle("x");
  sine_graph->GetYaxis()->SetTitle("sin#left(x#right)");

  // Plot the graph
  sine_graph->Draw("AP");

  // Plot a sine-function on top
  TF1 *root_sin_fun = new TF1("root_sin_fun", "sin(x)", -M_PI, M_PI);
  root_sin_fun->SetLineWidth(3); // In pixel-units
  root_sin_fun->SetLineColor(2);
  root_sin_fun->Draw("same");

  // Draw a line, so we can see easier, where 0 is
  TLine *tl1 = new TLine(x_min, 0, x_max, 0); 
  tl1->SetLineStyle(2);
  tl1->SetLineWidth(2); // In pixel-units
  tl1->Draw("same");
  
  // Output the image as a PNG file
  cc->SaveAs("erzeuger.png");
}
